/*
 * Copyright 1999-2101 Alibaba Group.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.alibaba.fastjson.parser;

import java.io.File;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.lang.reflect.TypeVariable;
import java.lang.reflect.WildcardType;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.net.Inet4Address;
import java.net.Inet6Address;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.URI;
import java.net.URL;
import java.nio.charset.Charset;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Deque;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.TimeZone;
import java.util.TreeMap;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.regex.Pattern;

import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONException;
import com.alibaba.fastjson.JSONObject;
import com.alibaba.fastjson.parser.deserializer.AppendableDeserializer;
import com.alibaba.fastjson.parser.deserializer.ArrayDeserializer;
import com.alibaba.fastjson.parser.deserializer.ByteDeserializer;
import com.alibaba.fastjson.parser.deserializer.CharacterDeserializer;
import com.alibaba.fastjson.parser.deserializer.CharsetDeserializer;
import com.alibaba.fastjson.parser.deserializer.ConcurrentHashMapDeserializer;
import com.alibaba.fastjson.parser.deserializer.DateDeserializer;
import com.alibaba.fastjson.parser.deserializer.DefaultObjectDeserializer;
import com.alibaba.fastjson.parser.deserializer.DequeDeserializer;
import com.alibaba.fastjson.parser.deserializer.EnumDeserializer;
import com.alibaba.fastjson.parser.deserializer.FileDeserializer;
import com.alibaba.fastjson.parser.deserializer.HashMapDeserializer;
import com.alibaba.fastjson.parser.deserializer.InetAddressDeserializer;
import com.alibaba.fastjson.parser.deserializer.InetSocketAddressDeserializer;
import com.alibaba.fastjson.parser.deserializer.IntegerDeserializer;
import com.alibaba.fastjson.parser.deserializer.JSONArrayDeserializer;
import com.alibaba.fastjson.parser.deserializer.JSONObjectDeserializer;
import com.alibaba.fastjson.parser.deserializer.JavaObjectDeserializer;
import com.alibaba.fastjson.parser.deserializer.LinkedHashMapDeserializer;
import com.alibaba.fastjson.parser.deserializer.LocaleDeserializer;
import com.alibaba.fastjson.parser.deserializer.LongDeserializer;
import com.alibaba.fastjson.parser.deserializer.ObjectDeserializer;
import com.alibaba.fastjson.parser.deserializer.PatternDeserializer;
import com.alibaba.fastjson.parser.deserializer.ShortDeserializer;
import com.alibaba.fastjson.parser.deserializer.SqlDateDeserializer;
import com.alibaba.fastjson.parser.deserializer.StringDeserializer;
import com.alibaba.fastjson.parser.deserializer.ThrowableDeserializer;
import com.alibaba.fastjson.parser.deserializer.TimeZoneDeserializer;
import com.alibaba.fastjson.parser.deserializer.TimestampDeserializer;
import com.alibaba.fastjson.parser.deserializer.TreeMapDeserializer;
import com.alibaba.fastjson.parser.deserializer.URIDeserializer;
import com.alibaba.fastjson.parser.deserializer.URLDeserializer;
import com.alibaba.fastjson.parser.deserializer.UUIDDeserializer;

/**
 * @author wenshao<szujobs@hotmail.com>
 */
@SuppressWarnings("rawtypes")
public class DefaultExtJSONParser extends DefaultJSONParser implements ExtJSONParser {

    private JavaBeanMapping                                mapping;

    private boolean                                        ignoreNotMatch     = true;

    private DefaultObjectDeserializer                      derializer         = new DefaultObjectDeserializer();

    private final Map<Class<?>, ObjectDeserializer>        derializers        = defaultDerializers;

    private final static Map<Class<?>, ObjectDeserializer> defaultDerializers = new HashMap<Class<?>, ObjectDeserializer>();
    static {
        defaultDerializers.put(java.sql.Timestamp.class, TimestampDeserializer.instance);
        defaultDerializers.put(java.sql.Date.class, SqlDateDeserializer.instance);
        defaultDerializers.put(java.util.Date.class, DateDeserializer.instance);

        defaultDerializers.put(JSONObject.class, JSONObjectDeserializer.instance);
        defaultDerializers.put(JSONArray.class, JSONArrayDeserializer.instance);

        defaultDerializers.put(Map.class, HashMapDeserializer.instance);
        defaultDerializers.put(HashMap.class, HashMapDeserializer.instance);
        defaultDerializers.put(LinkedHashMap.class, LinkedHashMapDeserializer.instance);
        defaultDerializers.put(TreeMap.class, TreeMapDeserializer.instance);
        defaultDerializers.put(ConcurrentMap.class, ConcurrentHashMapDeserializer.instance);
        defaultDerializers.put(ConcurrentHashMap.class, ConcurrentHashMapDeserializer.instance);
        defaultDerializers.put(Deque.class, DequeDeserializer.instance);
        defaultDerializers.put(ArrayDeque.class, DequeDeserializer.instance);

        defaultDerializers.put(Object.class, JavaObjectDeserializer.instance);
        defaultDerializers.put(String.class, StringDeserializer.instance);
        defaultDerializers.put(char.class, CharacterDeserializer.instance);
        defaultDerializers.put(Character.class, CharacterDeserializer.instance);
        defaultDerializers.put(byte.class, ByteDeserializer.instance);
        defaultDerializers.put(Byte.class, ByteDeserializer.instance);
        defaultDerializers.put(short.class, ShortDeserializer.instance);
        defaultDerializers.put(Short.class, ShortDeserializer.instance);
        defaultDerializers.put(int.class, IntegerDeserializer.instance);
        defaultDerializers.put(Integer.class, IntegerDeserializer.instance);
        defaultDerializers.put(long.class, LongDeserializer.instance);
        defaultDerializers.put(Long.class, LongDeserializer.instance);
        
        defaultDerializers.put(UUID.class, UUIDDeserializer.instance);
        defaultDerializers.put(TimeZone.class, TimeZoneDeserializer.instance);
        defaultDerializers.put(Locale.class, LocaleDeserializer.instance);
        defaultDerializers.put(InetAddress.class, InetAddressDeserializer.instance);
        defaultDerializers.put(Inet4Address.class, InetAddressDeserializer.instance);
        defaultDerializers.put(Inet6Address.class, InetAddressDeserializer.instance);
        defaultDerializers.put(InetSocketAddress.class, InetSocketAddressDeserializer.instance);
        defaultDerializers.put(File.class, FileDeserializer.instance);
        defaultDerializers.put(URI.class, URIDeserializer.instance);
        defaultDerializers.put(URL.class, URLDeserializer.instance);
        defaultDerializers.put(Appendable.class, AppendableDeserializer.instance);
        defaultDerializers.put(StringBuilder.class, AppendableDeserializer.instance);
        defaultDerializers.put(StringBuffer.class, AppendableDeserializer.instance);
        defaultDerializers.put(Pattern.class, PatternDeserializer.instance);
        defaultDerializers.put(Charset.class, CharsetDeserializer.instance);
    }

    private final static Set<Class<?>>                     primitiveClasses   = new HashSet<Class<?>>();
    static {
        primitiveClasses.add(boolean.class);
        primitiveClasses.add(byte.class);
        primitiveClasses.add(short.class);
        primitiveClasses.add(int.class);
        primitiveClasses.add(long.class);
        primitiveClasses.add(float.class);
        primitiveClasses.add(double.class);

        primitiveClasses.add(Boolean.class);
        primitiveClasses.add(Byte.class);
        primitiveClasses.add(Short.class);
        primitiveClasses.add(Integer.class);
        primitiveClasses.add(Long.class);
        primitiveClasses.add(Float.class);
        primitiveClasses.add(Double.class);

        primitiveClasses.add(BigInteger.class);
        primitiveClasses.add(BigDecimal.class);
        primitiveClasses.add(String.class);
    }

    public DefaultExtJSONParser(String input){
        this(input, JavaBeanMapping.getGlobalInstance());
    }

    public DefaultExtJSONParser(String input, JavaBeanMapping mapping){
        super(input);
        this.mapping = mapping;
    }

    public DefaultExtJSONParser(String input, SymbolTable symbolTable){
        this(input, symbolTable, JavaBeanMapping.getGlobalInstance());
    }

    public DefaultExtJSONParser(String input, SymbolTable symbolTable, JavaBeanMapping mapping){
        super(input, symbolTable);
        this.mapping = mapping;
    }

    public DefaultExtJSONParser(String input, SymbolTable symbolTable, int features){
        this(input, symbolTable, features, JavaBeanMapping.getGlobalInstance());
    }

    public DefaultExtJSONParser(String input, SymbolTable symbolTable, int features, JavaBeanMapping mapping){
        super(input, symbolTable, features);
        this.mapping = mapping;
    }

    public JavaBeanMapping getMapping() {
        return mapping;
    }

    public void setMapping(JavaBeanMapping mapping) {
        this.mapping = mapping;
    }

    public boolean isIgnoreNotMatch() {
        return ignoreNotMatch;
    }

    public void setIgnoreNotMatch(boolean ignoreNotMatch) {
        this.ignoreNotMatch = ignoreNotMatch;
    }
    
    // compatible
    @SuppressWarnings("unchecked")
    public <T> T parseObject(Class<T> clazz) {
        return (T) parseObject((Type) clazz);
    }

    @SuppressWarnings("unchecked")
    public <T> T parseObject(Type type) {
        if (lexer.token() == JSONToken.NULL) {
            lexer.nextToken();
            return null;
        }

        ObjectDeserializer derializer = derializers.get(type);
        if (derializer != null) {
            return (T) derializer.deserialze(this, type);
        }

        try {
            if (type instanceof Class<?>) {
                Class<?> clazz = (Class<?>) type;
                
                if (clazz.isEnum()) {
                    return (T) EnumDeserializer.instance.deserialze(this, clazz);
                } else if (clazz.isArray()) {
                    return (T) ArrayDeserializer.instance.deserialze(this, clazz);
                } else if (primitiveClasses.contains(clazz)) {
                    Object jsonValue = parse();
                    return (T) TypeUtils.cast(jsonValue, clazz, mapping);
                } else if (Throwable.class.isAssignableFrom(clazz)) {
                    return (T) ThrowableDeserializer.instance.deserialze(this, clazz);
                } else if (Appendable.class.isAssignableFrom(clazz)) {
                    return (T) AppendableDeserializer.instance.deserialze(this, clazz);
                }
            }
            
            return (T)this.derializer.deserialze(this, type);
        } catch (JSONException e) {
            throw e;
        } catch (Throwable e) {
            throw new JSONException(e.getMessage(), e);
        }
    }

    public <T> List<T> parseArray(Class<T> clazz) {
        List<T> array = new ArrayList<T>();
        parseArray(clazz, array);
        return array;
    }

    @SuppressWarnings({ "unchecked" })
    public void parseArray(Class<?> clazz, Collection array) {
        accept(JSONToken.LBRACKET);

        for (;;) {
            if (isEnabled(Feature.AllowArbitraryCommas)) {
                while (lexer.token() == JSONToken.COMMA) {
                    lexer.nextToken();
                    continue;
                }
            }

            if (lexer.token() == JSONToken.RBRACKET) {
                break;
            }

            Object val = parseObject(clazz);
            array.add(val);

            if (lexer.token() == JSONToken.COMMA) {
                lexer.nextToken();
                continue;
            }
        }

        accept(JSONToken.RBRACKET);
    }

    public void parseObject(Object object) {
        derializer.parseObject(this, object);
    }

    public Object parseArrayWithType(Type collectionType) {
        if (lexer.token() == JSONToken.NULL) {
            lexer.nextToken();
            return null;
        }

        Type actualTypeArgument = ((ParameterizedType) collectionType).getActualTypeArguments()[0];

        if (actualTypeArgument instanceof Class) {
            List<Object> array = new ArrayList<Object>();
            this.parseArray((Class) actualTypeArgument, array);
            return array;
        }

        if (actualTypeArgument instanceof WildcardType) {
            WildcardType wildcardType = (WildcardType) actualTypeArgument;

            if (wildcardType.getUpperBounds().length == 1) {
                Type upperBoundType = wildcardType.getUpperBounds()[0];
                if (upperBoundType instanceof Class) {
                    if (Object.class.equals(upperBoundType)) {
                        if (wildcardType.getLowerBounds().length == 0) {
                            // Collection<?>
                            return parse();
                        } else {
                            throw new JSONException("not support type : " + collectionType);
                        }
                    }

                    List<Object> array = new ArrayList<Object>();
                    this.parseArray((Class) upperBoundType, array);
                    return array;
                }
            }

            // throw new JSONException("not support type : " +
            // collectionType);return parse();
        }

        if (actualTypeArgument instanceof TypeVariable) {
            TypeVariable typeVariable = (TypeVariable) actualTypeArgument;
            Type[] bounds = typeVariable.getBounds();

            if (bounds.length != 1) {
                throw new JSONException("not support : " + typeVariable);
            }

            Type boundType = bounds[0];
            if (boundType instanceof Class) {
                List<Object> array = new ArrayList<Object>();
                this.parseArray((Class) boundType, array);
                return array;
            }
        }

        if (actualTypeArgument instanceof ParameterizedType) {
            ParameterizedType parameterizedType = (ParameterizedType) actualTypeArgument;

            List<Object> array = new ArrayList<Object>();
            this.parseArray((Class) parameterizedType.getRawType(), array);
            return array;
        }

        throw new JSONException("TODO : " + collectionType);
    }
}
